class Foreground:
    def color(s, color):
        if color < 31 or color > 37: return
        s.val["fore"] = f"\033[{str(color)}m"
        return s
    
    @property
    def black(s):
        s.color(30)
        return s
    
    @property
    def red(s):
        s.color(31)
        return s
    
    @property
    def green(s):
        s.color(32)
        return s
    
    @property
    def yellow(s):
        s.color(33)
        return s
    
    @property
    def blue(s):
        s.color(34)
        return s
    
    @property
    def magenta(s):
        s.color(35)
        return s
    
    @property
    def cyan(s):
        s.color(36)
        return s
    
    @property
    def white(s):
        s.color(37)
        return s


from .Attributes import Attributes

def ansi(*names):
    combined = ""
    for name in names:
        if not name in Attributes:
            continue
        combined += Attributes[name]
    return combined


from .Background import Background
from .Foreground import Foreground
from .Styles import Styles
from .All import All

class Color(Foreground, Background, Styles):
    def __init__(s, text="", *nests, after="", back="", fore="", style="", space_b="", space_a="", pad_b="", pad_a="", applied=lambda x, y: x, ALL=False, sp=" ", sp_a=True):
        s.sp = sp
        if isinstance(ALL, All):
            s.val = ALL
        else:
            s.val = All(text, back, fore, style, space_b, space_a, pad_b, pad_a, applied)
        s(s.val["text"], *nests, after=after, sp_a=sp_a)

    def __str__(s):
        return s.val["all"]
    
    def __repr__(s):
        return s.val["text"]
    
    def __call__(s, text, *nests, after="", sp_a=True):
        s.val["text"] = text
        s.nest(*nests, sp_a=sp_a)
        s.val["text", False] = str(after)
        return s
    
    def __iter__(s):
        return iter(str(s.val["all"]))

    def __add__(s, other):
        return s.val["all"] + other
    
    def __mul__(s, times):
        return s.val["all"] * times

    def after(s, string, *nests, after="", sp_a=True):
        s.val["text", False] = s.sp*sp_a + str(string)
        s.nest(*nests)
        s.val["text", False] = str(after)
        return s
    
    def before(s, string, *nests, after="", sp_a=True):
        original = s.sp*sp_a + s.val["text"]
        s.val["text"] = string
        s.nest(*nests)
        s.val["text", False] = str(after) + original
        return s
    
    def nest(s, *nests, sp_a=True):
        for i, nest in enumerate(nests):
            if i % 2 == 0:
                s.val["text"] += s.sp*sp_a + str(All.esc) + nest + s.val.attr()
            else:
                s.val["text"] += s.sp*sp_a + nest
        return s
    
    def space(s, spaces, before=True, after=True, clear=True):
        if before: s.val['space_b', clear] = spaces * " "
        if after: s.val['space_a', clear] = spaces * " "
        return s
    
    def space_b(s, spaces, clear=True):
        s.space(spaces, before=True, clear=clear)
        return s
    
    def space_a(s, spaces, clear=True):
        s.space(spaces, before=False, clear=clear)
        return s
    
    def pad(s, spaces, before=True, after=True, clear=True):
        if after: s.val["pad_b", clear] = spaces * " "
        if before: s.val["pad_a", clear] = spaces * " "
        return s
    
    def pad_b(s, spaces, clear=True):
        s.pad(spaces, before=False, clear=clear)
        return s

    def pad_a(s, spaces, clear=True):
        s.pad(spaces, after=False, clear=clear)
        return s

    def clear(s, text=False, back=False, fore=False, style=False, space_b=False, space_a=False, pad_b=False, pad_a=False, applied=False):
        s.val.clear(text, back, fore, style, space_b, space_a, pad_b, pad_a, applied)
        return s
    
    def clear_space(s, b=True, a=True):
        if b: s.clear(space_b=True)
        if a: s.clear(space_a=True)
        return s
    
    def clear_pad(s, l=True, r=True):
        if l: s.clear(pad_b=True)
        if r: s.clear(pad_a=True)
        return s
    
    def clear_text(s):
        s.clear(text=True)
        return s
    
    def clear_back(s):
        s.clear(back=True)
        return s

    def clear_fore(s):
        s.clear(fore=True)
        return s
    
    def clear_applied(s):
        s.clear(applied=True)
    
    def clear_style(s):
        s.clear(style=True)
        return s

    def new_sp(s, text):
        s.sp = text
        return s
    
    def remove_sp(s):
        s.sp = ""
        return s
    
    def reset(s):
        s.val.clear(True, True, True, True, True, True, True, True, True)
        return s

    def new(s, new_text="", text=False, back=True, fore=True, style=True, space_b=True, space_a=True, pad_b=True, pad_a=True):
        all = All(s.val["text"] if text else new_text,
            s.val["back"] if back else "",
            s.val["fore"] if fore else "",
            s.val["style"] if style else "",
            s.val["space_b"] if space_b else "",
            s.val["space_a"] if space_a else "",
            s.val["pad_b"] if pad_b else "",
            s.val["pad_a"] if pad_a else "")
        return Color(ALL=all)
    
    def apply(s, func):
        s.val["applied"] = func
        return s
    
    def print(s, *args, **kwargs):
        print(s.val["all"], *args, **kwargs)

    @property
    def ansi(s):
        return f"{s.val['back']}{s.val['fore']}{s.val['style']}"
    
    def freeze(s):
        return Static(s.val.copy(), s.sp)
    
    @property
    def text(s):
        return s.val["text"]

class Static:
    def __init__(s, all, sp):
        s.val = all
        s.sp = sp
    
    def __str__(s):
        return ""
        
    def __call__(s, text, *nests, after=""):
        return Color(ALL=s.val.copy(), sp=s.sp)(text, *nests, after=after)


class Background:
    def bg_color(s, color):
        if color < 40 or color > 47: return
        s.val["back"] = f"\033[{str(color)}m"
        return s

    @property
    def bg_black(s):
        s.bg_color(40)
        return s
    
    @property
    def bg_red(s):
        s.bg_color(41)
        return s
    
    @property
    def bg_green(s):
        s.bg_color(42)
        return s

    @property
    def bg_yellow(s):
        s.bg_color(43)
        return s

    @property
    def bg_blue(s):
        s.bg_color(44)
        return s

    @property
    def bg_magenta(s):
        s.bg_color(45)
        return s

    @property
    def bg_cyan(s):
        s.bg_color(46)
        return s

    @property
    def bg_white(s):
        s.bg_color(47)
        return s



class Styles:
    def style(s, style):
        s.val["style", False] = f"\033[{str(style)}m"
        return s
    
    @property
    def bold(s):
        s.style(1)
        return s
    
    @property
    def dim(s):
        s.style(2)
        return s
        
    @property
    def underline(s):
        s.style(4)
        return s
    
    @property
    def blink(s):
        s.style(5)
        return s
    
    @property
    def rapid_blink(s):
        s.style(6)
        return s
    
    @property
    def inverse(s):
        s.style(7)
        return s
    
    @property
    def hidden(s):
        s.style(8)
        return s
    
    @property
    def strikethrough(s):
        s.style(9)
        return s


class All:
    esc = "\033[0m"
    next_key = 1

    def __init__(s, text, back, fore, style, space_b, space_a, pad_b, pad_a, applied):
        s.exclusive_key = f"C[[CLASS_ALL_APPLY_STYLES_HERE_{str(All.next_key)}]]C"
        All.next_key += 1
        s.text = text
        s.back = back
        s.fore = fore
        s.style = style
        s.space_b = space_b
        s.space_a = space_a
        s.pad_b = pad_b
        s.pad_a = pad_a
        s.applied = applied
        s.update()

    def __str__(s):
        return s.all
        
    def __getitem__(s, key):
        return getattr(s, key)
    
    def __setitem__(s, key, value):
        if isinstance(key, tuple): 
            act_key, should_override = key
        else:
            act_key, should_override = key, True
        if act_key == "all" or not hasattr(s, act_key): 
            return
        if should_override:
            setattr(s, act_key, value)
        else:
            setattr(s, act_key, str(getattr(s, act_key)) + str(value))
        s.update()

    def __delitem__(s, key):
        setattr(s, key, "")
    
    def __attr(s):
        return s.back + s.fore + s.style
    
    def attr(s):
        return s.exclusive_key
    
    def copy(s):
        return All(s.text, s.back, s.fore, s.style, s.space_b, s.space_a, s.pad_b, s.pad_a, s.applied)

    def clear(s, text, back, fore, style, space_b, space_a, pad_b, pad_a, applied):
        if text: s.text = ""
        if back: s.back = ""
        if fore: s.fore = ""
        if style: s.style = ""
        if space_b: s.space_b = ""
        if space_a: s.space_a = ""
        if pad_b: s.pad_b = ""
        if pad_a: s.pad_a = ""
        if applied: s.applied = lambda string, key: string
        s.update()
    
    def update(s):
        s.all = (
            str(s.space_b) + 
            str(s.back) + str(s.fore) + str(s.style) + 
            str(s.pad_b) +
            str(s.applied(s.text, s.exclusive_key)).replace(s.exclusive_key, s.__attr()) + 
            str(s.pad_a) +
            str(All.esc) + 
            str(s.space_a)
        )
